/*
    thumbnail

    v0.01 - initial

    -on timeshift playback, fill in gaps in .ss file

*/
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <tap.h>
#include "exTap.h"
#include "tf5800.h"
#include "thumbnail.h"

#ifdef _WIN32
#else
#define ID_USERTAP 0xbdb02501
TAP_ID( ID_USERTAP );
TAP_PROGRAM_NAME("thumbnail v0.01");
TAP_AUTHOR_NAME("bdb");
TAP_DESCRIPTION("a thumbnail tap");
TAP_ETCINFO(__DATE__);
#endif

//----------------------------------------------------------------------------
int five_seconds_enabled = 1;
int show_thumbnail_enabled = 0;
int show_memory_enabled = 0;
dword five_seconds = 0;
void five_second_update();
void dump_memory();
void setupGraphics();
void thumbnail_status();
void thumbnail_display();

//----------------------------------------------------------------------------
//13.40
type_ts_struct1 *fw_ts_struct1; // = (type_ts_struct1 *)(fw_gp-0x611c);
type_ts_struct2 *fw_ts_struct2; // = (type_ts_struct2 *)(fw_gp-0x787c);

//
//the block number
//- 60x60=3600 4 bytes per second - block offset
//
dword *fw_tshift_block_offset = (dword *)0x81b76230;

//void fw_grab_screen_image(int mainSub, word *buf, int width, int height);
void *fw_grab_screen_image = (void *)0x8017129c;

////int fw_fseek(int handle, int offset, int origin);
//dword fw_fseek = 0x80143ae4;
//
////int fw_fread(void *buf, int size, int num, int handle);
//dword fw_fread = 0x80143744;
//
////int fw_fwrite(void *buf, int size, int num, int handle);
//dword fw_fwrite = 0x801438e0;

//----------------------------------------------------------------------------
//menu stuff
char *tap_name;
type_menu main_menu = {
    0,
    0,
    {0},
    0,
    (char *)__tap_program_name__,
    {
        {5, "xxx",    },

        {0, "Thumbnail Status", },
        {1, "Disable Thumbnail Update", },
        {2, "Show Current Thumbnail", },
        {3, "Show Memory Usage", },

        {8, "Exit Tap", },
    },
};

//----------------------------------------------------------------------------
//
//tsr_commander stuff
//
#ifdef _WIN32
#else
#include "TSRCommander.inc"

bool TSRCommanderExitTAP()
{
    close();
    return FALSE;
}

void TSRCommanderConfigDialog()
{
    ShowMenu(&main_menu);
}

int is_tsr_commander_running(void)
{
    int i;
    type_TAPProcess *TAPProcess;

    for (i=0; i<16; i++) {
        TAPProcess = (type_TAPProcess *)(fw_tap_table + (i*32));
        if (TAPProcess->firmwaregp) {
            if (TAPProcess->header) {
                if (TAPProcess->header->id == 0x807a10b2) {  //tsr_commander
                    return 1;
                }
                if (TAPProcess->header->id == 0x81001021) {  //tap_commander
                    return 1;
                }
            }
        }
    }
    return 0;
}
#endif

//----------------------------------------------------------------------------
int TAP_Main(void)
{
    int ret;

    tap_name =  (char *)__tap_program_name__;
    sprintf(strLog, "Opening %s", tap_name);
    log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_SHORT);

    //determine support for different firmwares
    ret = exTAP_Init();
    if (ret || (fw_sysid != 456) || (_appl_version != 0x1340)) {
        sprintf(strLog, "Unsupported platform - %d:%04x", fw_sysid, _appl_version);
        log_message(LOG_SCREEN_WAIT|LOG_SERIAL, strLog, DELAY_MEDIUM);
        return 0;
    }

    //13.40
    fw_ts_struct1 = (type_ts_struct1 *)(fw_gp-0x611c);
    fw_ts_struct2 = (type_ts_struct2 *)(fw_gp-0x787c);

    main_menu.visible = 0;
    setupGraphics();
    TSRCommanderInit(0, TRUE);
    return 1;                 //0=application mode, 1=tsr mode
}

int close(void)
{
    DeleteMenu(&main_menu);

    sprintf(strLog, "Closing  %s", tap_name);
    log_message(LOG_SERIAL|LOG_SCREEN_WAIT, strLog, DELAY_SHORT);
    if (message_rgn) TAP_Osd_Delete(message_rgn);
    if (OpLog) TAP_Hdd_Fclose(OpLog);
    TSRCommanderErase();
    TAP_Print("\nClosing %s", tap_name);
    TAP_Exit();
    return 0;
}
//----------------------------------------------------------------------------
dword TAP_EventHandler(word event, dword param1, dword param2)
{
    dword state, subState;
    int tick;

    switch (event) {
        case EVT_IDLE:
            tick = TAP_GetTick();
            //
            //message window
            //
            if (message_showing) {
                if (tick > message_showing) {
                    TAP_Osd_Delete(message_rgn);
                    message_rgn = 0;
                    message_showing = 0;
                }
            }

            //
            //housekeeping
            //
            if (five_seconds_enabled) {
                if (TAP_GetTick() > five_seconds) {
                    five_second_update();
                    five_seconds = TAP_GetTick() + (5*100);
                }
            }

            TSRCommanderWork();
            break;

        case EVT_KEY:
            TAP_GetState(&state, &subState);
//            TAP_Print("\nEVT_KEY, event=%x, param1=%x, param2=%x", event, param1, param2);
//            TAP_Print("\nstate = %s, subState = %s", get_str(state_str, state), get_str(substate_str, subState));
            if (main_menu.visible) {
                param1 = menu_handler(&main_menu, param1);
            }
            else {
                if ((state == STATE_Normal) && (subState == SUBSTATE_Normal)) {
                    param1 = key_handler(param1, param2);
                }
            }
            break;
        default:
            break;
    }

    return param1;
}


//----------------------------------------------------------------------------
int menu_handler(type_menu *menu, dword param1)
{
    int selection;

    switch (param1) {
        case RKEY_Ok:
            selection = TAP_Win_GetSelection(&menu->win);
            DeleteMenu(menu);
            switch (selection) {
                case 0:
                    thumbnail_status();
                    return 0;
                case 1:
                    five_seconds_enabled ^= 1;
                    if (five_seconds_enabled) {
                        strcpy(menu->item[2].str, "Disable Thumbnail Update");
                    }
                    else {
                        strcpy(menu->item[2].str, "Enable Thumbnail Update");
                    }
                    return 0;
                case 2:
                    show_thumbnail_enabled ^= 1;
                    if (show_thumbnail_enabled) {
                        strcpy(menu->item[3].str, "Clear Current Thumbnail");
                        thumbnail_display();
                    }
                    else {
                        strcpy(menu->item[3].str, "Show Current Thumbnail");
                    }
                    return 0;
                case 3:
                    show_memory_enabled ^= 1;
                    if (show_memory_enabled) {
                        strcpy(menu->item[4].str, "Clear Memory Usage");
                        dump_memory();
                    }
                    else {
                        strcpy(menu->item[4].str, "Show Memory Usage");
                    }
                    return 0;
                case 4:
                    close();
                    return 0;
            }
            break;
        case RKEY_Recall:
            DeleteMenu(menu);
            return param1;
        case RKEY_Exit:
            DeleteMenu(menu);
            return 0;
        default:
            TAP_Win_Action(&menu->win, param1);
            return 0;
    }
    return param1;
}

int key_handler(dword param1, dword param2)
{
    switch(param1) {
        //only if tsrcommander is not running
        case RKEY_Recall:
            if (!is_tsr_commander_running()) {
                ShowMenu(&main_menu);
                return 0;
            }
            break;
    }
    return param1;
}

//-----------------------------------------------------------------------------
//
//print multiline message to OSD
//  500=vertical position(0-576)
//  RGB(3,5,10)=background colour(dark blue)
//  200=hold time(2 seconds)
//  0=blocking
//
void oprintf(const void *fmt, ...)
{
    va_list argp;

    va_start(argp, fmt);
    vsprintf(strLog, fmt, argp);
//    TAP_Print("\n%s", strLog);
    ShowMessage_y(strLog, 500, RGB(3,5,10), 500, 0);
    va_end(argp);

    return;
}

//----------------------------------------------------------------------------
word *_vAddr[576];      //osd buffer, access with _vAddr[y]+x

void setupGraphics()
{
   int i;
   TYPE_OsdBaseInfo osdBaseInfo;

   TAP_Osd_GetBaseInfo(&osdBaseInfo);

   // set vertical address
   for(i=0; i<576; i+=2) {
       _vAddr[i] = osdBaseInfo.eAddr + 720*i;
       _vAddr[i+1] = osdBaseInfo.oAddr + 720*i;
   }
}

int bitblt(int x, int y, int width, int height, word *bitmap)
{
    int i,j;

    for(i=0; i<width; i++) {
        for(j=0; j<height; j++) {
            *(_vAddr[y+j]+x+i) = bitmap[width*j+i];
        }
    }

    return 0;
}

int bitclr(int x, int y, int width, int height)
{
    int i,j;

    for(i=0; i<width; i++) {
        for(j=0; j<height; j++) {
            *(_vAddr[y+j]+x+i) = 0;
        }
    }

    return 0;
}
//----------------------------------------------------------------------------
void dump_memory()
{
    dword heapSize, freeHeapSize, availHeapSize;

    TAP_MemInfo(&heapSize, &freeHeapSize, &availHeapSize);
    oprintf("heapSize=%d        freeHeapSize=%d        availHeapSize=%d", heapSize, freeHeapSize, availHeapSize);
}

void five_second_update()
{
    if (show_memory_enabled) dump_memory();
    thumbnail_save();
//    if (show_thumbnail_enabled) thumbnail_display();
}


//----------------------------------------------------------------------------
//
//.ss file seems to be a compressed bitmap of 144x115 pixels
//it is extracted from the screen buffer using gba(1,x,&result)
//so can only be generated when actually watching the live signal
//i.e. not when timeshifting, non-viewed recording
//it is only kept while the timeshift buffer is active
//0x8200 bytes per frame
//
word ss_buf[0x4100];        //144x115 x16bit

//
//if in timeshift mode, check if a thumbnail exists
//if not, save one
//
void thumbnail_status()
{
    int ret;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    TYPE_File *ss_handle = NULL;
    byte *ss_valid_table;
    int ss_valid;
    int x,y;
    int ss_buffer_length;
    int ss_write_position;
    int ss_read_offset;
    int ss_read_position;

    TAP_Hdd_GetPlayInfo(&pi);
    if (pi.playMode == PLAYMODE_TempPlaying) {
        //get the .ss file handle
        ss_handle = (TYPE_File *)(fw_ts_struct2->ss_handle);

        ss_buffer_length = fw_ts_struct1->ts_buffer_length;
        ss_write_position = fw_ts_struct1->ts_write_position;
        ss_read_offset = fw_ts_struct1->ts_read_offset;

        ss_read_position = ss_write_position+ss_read_offset;
        ss_read_position = (ss_read_position+5)/10;

//        //check if we already have this thumbnail
//        ss_valid_table  = (byte *)fw_ts_struct1->ss_valid_table;
//        ss_valid = ss_valid_table[ss_read_position];

        TAP_Print("\nThumbnail Status");
        TAP_Print("\nTimeshift Play Position: -%02d:%02d", -ss_read_offset/60, -ss_read_offset%60);
        TAP_Print("\nTimeshift Buffer Length:  %02d:%02d", ss_buffer_length/60, ss_buffer_length%60);

        oprintf("Timeshift: Position: -%02d:%02d, Length: %02d:%02d", -ss_read_offset/60, -ss_read_offset%60, ss_buffer_length/60, ss_buffer_length%60);
    }
    else {
        oprintf("Not in Timeshift mode");
    }
}

int thumbnail_save()
{
    int i;
    int ret;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    TYPE_File *ss_handle = NULL;
    static TYPE_File *ss_handle2 = NULL;
    static TYPE_File *current_file = NULL;
    TYPE_File file;
    byte *ss_valid_table;
    int ss_valid;
    int x,y;
    int ss_write_position;
    int ss_read_offset;
    int ss_read_position;

    char filename[0x40];
    int  file_length_seconds;
    int  file_length_blocks;
    int  blocks_per_10seconds;
    static int  total_thumbnails = 0;
    int  current_thumbnail;
    static dword *thumbnail_index = NULL;
    static int index_offset = 0;
    static int thumbnail_index_dirty = 0;


    TAP_Hdd_GetPlayInfo(&pi);
    if (pi.playMode == PLAYMODE_TempPlaying) {
        //get the .ss file handle
        ss_handle = (TYPE_File *)(fw_ts_struct2->ss_handle);

        ss_write_position = fw_ts_struct1->ts_write_position;
        ss_read_offset = fw_ts_struct1->ts_read_offset;

        ss_read_position = ss_write_position+ss_read_offset;
        ss_read_position = (ss_read_position+5)/10;

        //check if we already have this thumbnail
        ss_valid_table  = (byte *)fw_ts_struct1->ss_valid_table;
        ss_valid = ss_valid_table[ss_read_position];
        if (ss_valid == 0) {
            //grab a new thumbnail
            ret = exTAP_CallFirmware(fw_grab_screen_image, (dword)CHANNEL_Main, (dword)&ss_buf[0], 144, 115,  0,0,0,0,0,0,0);
            //save it
            ret = TAP_Hdd_Fseek(ss_handle, 0x4100*sizeof(word)*ss_read_position, SEEK_SET);
            ret = TAP_Hdd_Fwrite(&ss_buf[0], 0x4100, sizeof(word), ss_handle);
            //update the table
            ss_valid_table[ss_read_position] = 1;

            if (show_thumbnail_enabled) {
                x = 0;
                y = 0;
                bitblt(50+(150*x), 50+(120*y), 144, 115, &ss_buf[0]);
            }
        }
        else {
//            TAP_Print("\ngot - show the thumbnail");
//            //debug - show the image
//
            if (show_thumbnail_enabled) {
                //read 33k (0x8200) (16 bits per pixel) of .ss file
                ret = TAP_Hdd_Fseek(ss_handle, 0x4100*sizeof(word)*ss_read_position, SEEK_SET);
                ret = TAP_Hdd_Fread(&ss_buf[0], 0x4100, sizeof(word), ss_handle);
                x = 0;
                y = 0;
                bitblt(50+(150*x), 50+(120*y), 144, 115, &ss_buf[0]);
            }
        }
    }

    return 0;
}

void thumbnail_display()
{
    int ret;
    TYPE_PlayInfo pi;        //overflows!
    byte dummy2[0x80];
    TYPE_File *ss_handle = NULL;
    byte *ss_valid_table;
    int ss_valid;
    int x,y;
    int ss_write_position;
    int ss_read_offset;
    int ss_read_position;


    TAP_Hdd_GetPlayInfo(&pi);
    if (pi.playMode == PLAYMODE_TempPlaying) {
        ss_write_position = fw_ts_struct1->ts_write_position;
        ss_read_offset = fw_ts_struct1->ts_read_offset;

        ss_read_position = ss_write_position+ss_read_offset;
        ss_read_position = (ss_read_position+5)/10;

        ss_valid_table  = (byte *)fw_ts_struct1->ss_valid_table;
        ss_valid = ss_valid_table[ss_read_position];
        if (ss_valid) {
            //read 33k (0x8200) (16 bits per pixel) of .ss file
            ret = TAP_Hdd_Fseek(ss_handle, 0x4100*sizeof(word)*ss_read_position, SEEK_SET);
            ret = TAP_Hdd_Fread(&ss_buf[0], 0x4100, sizeof(word), ss_handle);
            x = 0;
            y = 0;
            bitblt(50+(150*x), 50+(120*y), 144, 115, &ss_buf[0]);
        }
        else {
            x = 0;
            y = 0;
            bitclr(50+(150*x), 50+(120*y), 144, 115);
        }
    }
    else {
        x = 0;
        y = 0;
        bitclr(50+(150*x), 50+(120*y), 144, 115);
    }
}

